/*
 * Copyright (c) 1998-2006 by Winfried Bruns and Robert Koch
 */

Int divisors (Int **vec, int dim, Int **loc, Int **traf, Int *lcm)
{
   int i, j;

   for (i=1; i<=dim; ++i)
   {
      for (j=1; j<=dim; ++j)
      {
         loc[i][j] = vec[i][j];
         traf[i][j] = 0;
      }
      traf[i][i] = 1;
   }
   diag (loc, traf, dim);
   loc[0][0] = *lcm = 1;
   for (j=1; j<=dim; ++j)
   {
      loc[0][0] *= (loc[0][j] = Labs (loc[j][j]));
      *lcm = Lcm (*lcm, loc[0][j]);
   }
   return (loc[0][0]);
}

void points (Int **vec, Int *neu, Int **s, Int *d, Int *a, int dim, Int lcm, il **sol, il *sup, il **frei, il *hlist, int **hmat, int hom, int mode, int hsets, int *hsets_nxt, int *hsets_set, int *hsets_len)
{
   int i;

   for (i=1; i<=dim; ++i)
         scalmultmod (lcm/d[i], s[i], lcm, dim);
   count (vec, neu, s, d, a, 1, dim, lcm, sol, sup, frei, hlist, hmat, hom, mode, hsets, hsets_nxt, hsets_set, hsets_len);
}

void count (Int **vec, Int *neu, Int **s, Int *d, Int *a, int n, int dim, Int lcm, il **sol, il *sup, il **frei, il *hlist, int **hmat, int hom, int mode, int hsets, int *hsets_nxt, int *hsets_set, int *hsets_len)
{
   if (n <= dim)
      for (a[n]=0; a[n]<d[n]; ++a[n])
         count (vec, neu, s, d, a, n+1, dim, lcm, sol, sup, frei, hlist, hmat, hom, mode, hsets, hsets_nxt, hsets_set, hsets_len);
   else
      nextpt (vec, neu, s, a, dim, lcm, sol, sup, frei, hlist, hmat, hom, mode, hsets, hsets_nxt, hsets_set, hsets_len);
}

void nextpt (Int **vec, Int *neu, Int **s, Int *a, int dim, Int lcm, il **sol, il *sup, il **frei, il *hlist, int **hmat, int hom, int mode, int hsets, int *hsets_nxt, int *hsets_set, int *hsets_len)
{
   int i, j;

   for (j=1; j<=dim; ++j)
      s[0][j] = 0;
   for (i=1; i<=dim; ++i)
      for (j=1; j<=dim; ++j)
         s[0][j] = (s[0][j] + a[i] * s[i][j]) % lcm;
   for (j=1; j<=dim; ++j)
      if (s[0][j] < 0)
         s[0][j] += lcm;
   multvmv (s[0], dim, vec, dim, neu);
   scaldiv (lcm, neu, dim);
   if (hom)
      wrthmat (hlist, hmat, zerone (s[0], dim), neu[1], dim, hsets, hsets_nxt, hsets_set, hsets_len);
   reduce (sol, neu, dim, sup, frei, s[0], mode);
}

ul zerone (Int *coeff, int dim)
{
   int i;
   ul set;

   set = 0;
   for (i=dim; i>0; --i)
   {
      set <<= 1;
      if (coeff[i] != 0)
         set += 1;
   }
   return (set);
}

void wrthmat (il *hlist, int **hmat, ul coeff, Int Deg, int dim, int hsets, int *hsets_nxt, int *hsets_set, int *hsets_len)
{
  long a, b, c, i, deg; // wj {
  ul mask, temp;     // wj }

#ifdef INTEGER
#ifdef _Integer_h            // wj {
  deg = Deg.as_long();         // g++
#else
  deg = to_long (Deg);         // ntl
#endif                       // wj }
#else
  deg = Deg;
#endif

  //  printf("coeff = %d, deg = %d\n", coeff, deg);
  for (; hsets>=0; hsets=hsets_nxt[hsets])
    {
      //      printf("%d: %d %d %d\n", hsets, hsets_len[hsets], hsets_set[hsets], hsets_nxt[hsets]);
      mask = (1ul << hsets_len[hsets]) - 1;
      //      printf("mask = %d\n", mask);
      if ((coeff & mask & hsets_set[hsets])==(coeff & mask))
    {
      a=0;
      temp=coeff | (hsets_set[hsets] & mask);
      //      printf("%d\n", temp);
      for(i=0;i<hsets_len[hsets];++i)
        {
          if(temp & 1)
        {
          ++a;
        }
          temp>>=1;
        }
      for(b=0;temp;temp>>1)
        {
          if(temp & 1)
        {
          ++b;
        }
          temp>>=1;
        }
          c=dim-hsets_len[hsets]-b;
      //      printf("%d %d %d\n", a, b, c);
      temp=1;
      for(i=0;i<=c;++i)
        {
          if (i)
        {
          temp *= c-i+1;
          temp /= i;
        }
          //              printf("[%d][%d]+%d\n", a+b+i, a+b+i-deg, temp);
          hmat[a+b+i][a+b+i-deg] += temp;
        }
    }
    }
  //   for (; hlist; hlist=hlist->right)
  //      if (subset (coeff, hlist->set))
  //         ++(hmat[hlist->len][hlist->len-deg]);
}

int subset (ul cand, ul ref)
{
   if ((cand | ref) == ref)
      return (1);
   return (0);
}

int Subset (ul *cand, ul *ref, int bitc)
{
   if (bitc == 1)
   {
      if ((cand[0] | ref[0]) == ref[0])
         return (1);
      return (0);
   }
   for (int i=0; i<bitc; ++i)
      if ((cand[i] | ref[i]) != ref[i])
         return (0);
   return (1);
}

void bitset (ul *bitv, int pos)
{
   bitv[pos/LONG_BIT] |= (1ul << (pos%LONG_BIT));
}

void sumhvec (int **hmat, int *hvec, int dim)
{
   int i;

   for (i=0; i<=dim; ++i)
   {
      shift (hmat[i], i, dim);
      addvc (hmat[i], hvec, dim);
   }
}

void shift (int *a, int olddim, int newdim)
{
   int i, j;

   for (j=olddim; j<newdim; ++j)
      for (i=j+1; i>0; --i)
         a[i] -= a[i-1];
}

void addvc (int *a, int *b, int len)
{
   int i;

   for (i=0; i<=len; ++i)
      b[i] += a[i];
}

void setup (il **hreflist, il **hfullreflist, int **hmat, int *hvec, il **free, int dim)
{
   long i, j, k, set;
   il *ilptr[dim+1], *ilptr2;

   j = power (2, dim);
   for (i=0; i<j; ++i)
   {
      iisert (hfullreflist, free, 1);
      (*hfullreflist)->set = i;
      (*hfullreflist)->len = cntone ((*hfullreflist)->set);
      if ((*hfullreflist)->rightreset = (*hfullreflist)->right)
         (*hfullreflist)->right->leftreset = *hfullreflist;
   }
   for (i=0; i<=dim; ++i)
      ilptr[i] = NULL;
   j = power (2, dim-1);
   for (i=0; i<j; ++i)
   {
      set = 1;                     // [dim] %und [dim-1]% immer dabei
      set <<= (dim-1);
      set += i;
      k = cntone (set);
      iisert (&(ilptr[k]), free, 1);
      ilptr[k]->set = set;
      ilptr[k]->len = k;
      if (ilptr[k]->rightreset = ilptr[k]->right)
         ilptr[k]->right->leftreset = ilptr[k];
   }
   for (i=1; i<dim; ++i)
      icatchnx (&(ilptr[i+1]), &(ilptr[i]));
   *hreflist = ilptr[dim];
   for (i=0; i<=dim; ++i)
   {
      hvec[i] = 0;
      for (j=0; j<=dim; ++j)
         hmat[i][j] = 0;
   }
}

long power (int a, int n)
{
   int i;
   long b;

   b = 1;
   for (i=1; i<=n; ++i)
      b *= a;
   return (b);
}

int cntone (ul n)
{
   int i;

   i = 0;
   do
      i += n & 1;
   while (n >>= 1);
   return (i);
}

void reset (il **hlist, il **hreflist, il **hfullreflist)
{
   il *ilptr;

   if (*hlist == *hreflist || *hlist == *hfullreflist)
      *hlist = NULL;
   else
   {
      icatchn (hreflist, hlist);      // haengt hlist an hreflist
      for (ilptr=*hreflist; ilptr; ilptr=ilptr->rightreset)
      {
         ilptr->left = ilptr->leftreset;
         ilptr->right = ilptr->rightreset;
      }
      for (ilptr=(*hreflist)->leftreset; ilptr; ilptr=ilptr->leftreset)
      {
         ilptr->left = ilptr->leftreset;
         ilptr->right = ilptr->rightreset;
         *hreflist = ilptr;
      }
   }
}

void mark (int blcnt, int subblcnt, il **hreflist, il **hfullreflist, il **hlist, int dim, int bitc, pl *start, pl *end) // start/end no more
{
   int i, j;
   il *ilptr;

   // if (start == end)                           // attention!!!
   if (blcnt == 1)
      *hlist = *hfullreflist;
   else if (subblcnt == 1)
      *hlist = *hreflist;
   else
   {
      *hlist = *hreflist;
      *hreflist = NULL;

      for (ilptr=*hlist; ilptr;)
    if (!(ilptr->set & end->set) || inplist (ilptr->set, dim, bitc, start, end))
         {
       collsubsets (ilptr, dim, hreflist);
       // ATTENTION: ilptr == *hreflist would happen at beginning of a block; we avoid this
       iidlete (&ilptr, hreflist);
         }
         else
            ilptr = ilptr->right;
   }
}

void subblocks (pl **block, pl **subbl, int *numsubbl, int dim, int *oldpt, int *newpt)
{
   int action=0, i, numoldpt=0, numnewpt=0, hitnewpt;
   int delta[dim], numdelta, eps[dim], numeps;
   pl *plptr;

   --dim;                  // ATTENTION: [dim] is constant within block
   pdlete (block, &(subbl[1]));
   *numsubbl = 1;
   for (i=1; i<=dim; ++i)
      oldpt[i] = subbl[1]->ind[i];
   numoldpt = dim;
   while (*block)
   {
      for (plptr=*block; plptr; plptr=plptr->right)   // bilde neuen Subbl
      {
         // if (onenewpoint (plptr->ind, dim, oldpt, numoldpt, &hitnewpt))
         findverts (oldpt, numoldpt, newpt, 0, plptr, dim, delta, &numdelta, eps, &numeps);
         if (numeps == 1)
         {
            ++(*numsubbl);
            if (plptr == *block)
               *block = (*block)->right;
            pdlete (&plptr, &(subbl[*numsubbl]));
            ++numnewpt;
            // newpt[numnewpt] = hitnewpt;
            newpt[numnewpt] = eps[1];
            break;
         }
      }
      action = 1;
      while (action)               // teste nicht bearbeitete Simplices
      {
         action = 0;
         for (plptr=*block; plptr; )
         {
            findverts (oldpt, numoldpt, newpt, numnewpt, plptr, dim, delta, &numdelta, eps, &numeps);
            delta[0] = numdelta;
            if (!inpplist (delta, oldpt, numoldpt) && !inptlist (delta, numdelta, subbl[*numsubbl], dim))
            {
               action = 1;
               if (plptr == *block)
                  *block = (*block)->right;
               pdlete (&plptr, &(subbl[*numsubbl]));
               punion (newpt, &numnewpt, eps, &numeps);
            }
            else
               plptr=plptr->right;
         }
      }
      punion (oldpt, &numoldpt, newpt, &numnewpt);
   }
}

int onenewpoint (int *pt, int dim, int *oldpt, int noldpt, int *newpt)
{
   int i, hits=0, vec[2];

   vec[0] = 1;
   for (i=1; i<=dim; ++i)
   {
      vec[1] = pt[i];
      if (! inpplist (vec, oldpt, noldpt))
      {
         ++hits;
         *newpt = vec[1];
      }
      if (hits > 1)
         return (0);
   }
   if (hits == 1)
      return (1);
   return (0);
}

void findverts (int *oldpt, int numoldpt, int *newpt, int numnewpt, pl *simplex, int dim, int *delta, int *numdelta, int *eps, int *numeps)
{
   int i, vec[2];

   vec[0] = 1;
   *numdelta = *numeps = 0;
   simplex->set = 0;
   for (i=1; i<=dim; ++i)
   {
      vec[1] = simplex->ind[i];
      if (inpplist (vec, oldpt, numoldpt))
         delta[++(*numdelta)] = simplex->ind[i];
      else if (inpplist (vec, newpt, numnewpt))
      {
         delta[++(*numdelta)] = simplex->ind[i];
         simplex->set += power (2, i-1);
      }
      else
      {
         eps[++(*numeps)] = simplex->ind[i];
         simplex->set += power (2, i-1);
      }
   }
}

void punion (int *set, int *card, int *set2, int *card2) // geordn Vgg nach set
{
   while (*card2)
   {
      set[++(*card)] = set2[(*card2)--];
      lastsortvec (set, *card);
   }
}

int inplist (ul set, int dim, int bitc, pl *start, pl *end)
{
  int i;
  pl *plptr;
  ul bitv[bitc];

  ++stat1;
  for (i=0; i<bitc; ++i)
    bitv[i] = 0;
  for (i=0; i<dim; ++i)
    if (set & (1ul << i))
      {
    if (bitc == 1)
      *bitv |= (1ul << (end->ind[i+1]-1));
    else
      bitset (bitv, end->ind[i+1]-1);
      }
  for (plptr=start; plptr!=end; plptr=plptr->right)
    {
      ++stat2;
      if (bitc == 1)
    {
      if ((bitv[0] | plptr->bitv[0]) == plptr->bitv[0])
        return (1);
    }
      else if (Subset (bitv, plptr->bitv, bitc))
    return (1);
    }
  return (0);
}

int inptlist (int *pt, int npt, pl *plptr, int dim)
{
   pt[0] = npt;
   for (; plptr; plptr=plptr->right)
      if (inpplist (pt, plptr->ind, dim))
         return (1);
   return (0);
}

int inpplist (int *cand, int *obj, int dim)
{
   int i, j, ent, ok;

   ent = 1;
   for (i=1; i<=cand[0]; ++i)
   {
      ok = 0;
      for (j=ent; j<=dim; ++j)
         if (cand[i] > obj[j])
            continue;
         else if (cand[i] == obj[j])
         {
            ent = j + 1;
            ok = 1;
            break;
         }
      if (! ok)
         return (0);
   }
   return (1);
}

void collsubsets (il *ilptr, int dim, il **hlist)
{
   ul set=ilptr->set;

   for (ilptr=ilptr->right; ilptr;)
      if (subset (ilptr->set, set))
         iidlete (&ilptr, hlist);
      else
         ilptr = ilptr->right;
}

void collsupsets (il *ilptr, int dim, il **hlist)
{
   ul set=ilptr->set;

   for (ilptr=ilptr->right; ilptr;)
      if (supset (ilptr->set, set))
         iidlete (&ilptr, hlist);
      else
         ilptr = ilptr->right;
}

void marksubsets (il *ilptr)
{
   ul set=ilptr->set;

   for (; ilptr; ilptr=ilptr->right)
      if (subset (ilptr->set, set))
         ilptr->mark = 1;
}

void collmarksupsets (il *ilptr, int dim, il **hlist, il **free)
{
   ul set=ilptr->set;

   for (; ilptr; ilptr=ilptr->right)
      if ((ilptr->mark == 0) && supset (ilptr->set, set))
      {
         ilptr->mark = 1;
         iisert (hlist, free, dim);
         (*hlist)->set = ilptr->set;
         (*hlist)->len = ilptr->len;
      }
}

void delsubsets (il *beg, il **frhm)
{
   il *dum;

   for (dum=beg->right; dum;)
      if (subset (dum->set, beg->set))
      {
         idlete (&dum, frhm);
         if (dum == beg)
            break;
      }
      else
         dum = dum->right;
}

int supset (ul cand, ul ref)
{
   if ((cand | ref) == cand)
      return (1);
   return (0);
}

void hilbpoly (int *hvec, Int *hnum, Int *hden, Int *evec, Int **mat, int dim)
{
   int i, j;
   Int fac;

   fac = prod (1, dim);
   for (i=0; i<=dim; ++i)
   {
      hnum[i] = evec[i] = 0;
      hden[i] = fac;
      for (j=0; j<=dim; ++j)
         mat[i][j] = 0;
   }
   setevec (hvec, evec, dim, mat[0]);
   setpoly (mat, dim);
   for (i=0; i<dim; ++i)
      sscalmult (prod (i+1, dim) * parity (dim-1-i) * evec[dim-1-i], mat[i], dim);
   for (i=0; i<dim; ++i)
      addvtv (mat[i], hnum, dim);
   simcanc (hnum, hden, dim);
}

Int prod (int a, int b)
{
   int i;
   Int res;

   res = 1;
   for (i=a; i<=b; ++i)
      res *= i;
   return (res);
}

int parity (int a)
{
   if (a % 2)
      return (-1);
   return (1);
}

void setevec (int *hvec, Int *evec, int dim, Int *dum)
{
   int i, j;

   for (i=0; i<=dim; ++i)
      dum[i] = hvec[i];
   for (j=0; j<dim; ++j)
   {
      for (i=0; i<=dim; ++i)
         evec[j] += dum[i];
      evec[j] /= prod (1, j);
      diff (dum, dim);
   }
   for (i=0; i<=dim; ++i)
      dum[i] = 0;
}

void diff (Int *p, int len)
{
   int i;

   for (i=1; i<=len; ++i)
      p[i-1] = i * p[i];
}

void setpoly (Int **mat, int dim)
{
   int i, j;

   mat[0][0] = 1;
   for (i=1; i<=dim; ++i)
   {
      mat[i][0] = i * mat[i-1][0];
      for (j=1; j<dim; ++j)
         mat[i][j] = i * mat[i-1][j] + mat[i-1][j-1];
   }
}

void sscalmult (Int scal, Int *vec, int len)
{
   int i;

   for (i=0; i<=len; ++i)
      vec[i] *= scal;
}

void addvtv (Int *a, Int *b, int len)
{
   int i;

   for (i=0; i<=len; ++i)
      b[i] += a[i];
}

void simcanc (Int *a, Int *b, int len)
{
   int i;
   Int gcd;

   for (i=0; i<=len; ++i)
   {
      gcd = Labs (ggT (a[i], b[i]));
      a[i] /= gcd;
      b[i] /= gcd;
   }
}
